@@ -2,7 +2,7 @@
 /*                                                                          */
 /*  The FreeType project -- a free and portable quality TrueType renderer.  */
 /*                                                                          */
-/*  Copyright 1996-2000, 2003-2007, 2009-2014 by                            */
+/*  Copyright 1996-2018 by                                                  */
 /*  D. Turner, R.Wilhelm, and W. Lemberg                                    */
 /*                                                                          */
 /*                                                                          */
@@ -30,8 +30,7 @@
 #include FT_STROKER_H
 #include FT_SYNTHESIS_H
 #include FT_LCD_FILTER_H
-#include FT_CFF_DRIVER_H
-#include FT_TRUETYPE_DRIVER_H
+#include FT_DRIVER_H
 #include FT_MULTIPLE_MASTERS_H
 #include FT_SFNT_NAMES_H
 #include FT_TRUETYPE_IDS_H
@@ -44,8 +43,6 @@
 #define snprintf  _snprintf
 #endif
 
-#define N_CFF_HINTING_ENGINES  2
-
 
 #ifdef FT_DEBUG_AUTOFIT
   /* these variables, structures, and declarations are for  */
@@ -106,11 +103,11 @@
     int          Num;  /* glyph index */
     int          font_index;
 
-    FT_F26Dot6   scale;
+    int          scale;
     int          x_origin;
     int          y_origin;
 
-    FT_F26Dot6   scale_0;
+    int          scale_0;
     int          x_origin_0;
     int          y_origin_0;
 
@@ -131,15 +128,18 @@
     int          do_vert_hints;
     int          do_blue_hints;
     int          do_segment;
+    int          do_grid;
+    int          do_alt_colors;
 
     FT_LcdFilter lcd_filter;
-    double       gamma;
     const char*  header;
     char         header_buffer[BUFSIZE];
 
     FT_Stroker   stroker;
 
     unsigned int cff_hinting_engine;
+    unsigned int type1_hinting_engine;
+    unsigned int t1cid_hinting_engine;
     unsigned int tt_interpreter_versions[3];
     int          num_tt_interpreter_versions;
     int          tt_interpreter_version_idx;
@@ -153,6 +153,8 @@
     FT_UInt      current_axis;
     FT_UInt      used_num_axis;
 
+    int          no_named_instances;
+
   } GridStatusRec, *GridStatus;
 
   static GridStatusRec  status;
@@ -173,14 +175,17 @@
     st->do_vert_hints = 1;
     st->do_blue_hints = 1;
     st->do_segment    = 0;
+    st->do_grid       = 1;
+    st->do_alt_colors = 0;
 
     st->Num           = 0;
     st->lcd_filter    = FT_LCD_FILTER_DEFAULT;
-    st->gamma         = GAMMA;
     st->header        = NULL;
 
     st->mm            = NULL;
     st->current_axis  = 0;
+
+    st->no_named_instances = 0;
   }
 
 
@@ -191,7 +196,13 @@
     st->disp_width    = display->bitmap->width;
     st->disp_height   = display->bitmap->rows;
     st->disp_bitmap   = display->bitmap;
+  }
 
+
+  static void
+  grid_status_colors( GridStatus       st,
+                      FTDemo_Display*  display )
+  {
     st->axis_color    = grFindColor( display->bitmap,   0,   0,   0, 255 ); /* black       */
     st->grid_color    = grFindColor( display->bitmap, 192, 192, 192, 255 ); /* gray        */
     st->outline_color = grFindColor( display->bitmap, 255,   0,   0, 255 ); /* red         */
@@ -202,6 +213,22 @@
   }
 
 
+  static void
+  grid_status_alt_colors( GridStatus       st,
+                          FTDemo_Display*  display )
+  {
+    /* colours are adjusted for color-blind people, */
+    /* cf. http://jfly.iam.u-tokyo.ac.jp/color      */
+    st->axis_color    = grFindColor( display->bitmap,   0,   0,   0, 255 ); /* black          */
+    st->grid_color    = grFindColor( display->bitmap, 192, 192, 192, 255 ); /* gray           */
+    st->outline_color = grFindColor( display->bitmap, 230, 159,   0, 255 ); /* orange         */
+    st->on_color      = grFindColor( display->bitmap, 230, 159,   0, 255 ); /* orange         */
+    st->off_color     = grFindColor( display->bitmap,  86, 180, 233, 255 ); /* sky blue       */
+    st->segment_color = grFindColor( display->bitmap, 204, 121, 167,  64 ); /* reddish purple */
+    st->blue_color    = grFindColor( display->bitmap,   0, 114, 178,  64 ); /* blue           */
+  }
+
+
   static void
   grid_status_rescale_initial( GridStatus      st,
                                FTDemo_Handle*  handle )
@@ -265,9 +292,9 @@
   static void
   grid_status_draw_grid( GridStatus  st )
   {
-    int         x_org   = st->x_origin;
-    int         y_org   = st->y_origin;
-    FT_F26Dot6  xy_incr = st->scale;
+    int  x_org   = st->x_origin;
+    int  y_org   = st->y_origin;
+    int  xy_incr = st->scale;
 
 
     if ( xy_incr >= 2 )
@@ -304,8 +331,12 @@
 
   static void
   grid_hint_draw_segment( GridStatus     st,
+                          FT_Size        size,
                           AF_GlyphHints  hints )
   {
+    FT_Fixed  x_scale = size->metrics.x_scale;
+    FT_Fixed  y_scale = size->metrics.y_scale;
+
     FT_Int  dimension;
     int     x_org = st->x_origin;
     int     y_org = st->y_origin;
@@ -333,19 +364,24 @@
 
         if ( dimension == 0 ) /* AF_DIMENSION_HORZ is 0 */
         {
-          pos = x_org + ( ( offset * st->scale ) >> 6 );
+          offset = FT_MulFix( offset, x_scale );
+          pos    = x_org + ( ( offset * st->scale ) >> 6 );
           grFillVLine( st->disp_bitmap, pos, 0,
                        st->disp_height, st->segment_color );
         }
         else
         {
-          pos = y_org - ( ( offset * st->scale ) >> 6 );
+          offset = FT_MulFix( offset, y_scale );
+          pos    = y_org - ( ( offset * st->scale ) >> 6 );
 
           if ( is_blue )
           {
-            int  blue_pos = y_org - ( ( blue_offset * st->scale ) >> 6 );
+            int  blue_pos;
 
 
+            blue_offset = FT_MulFix( blue_offset, y_scale );
+            blue_pos    = y_org - ( ( blue_offset * st->scale ) >> 6 );
+
             if ( blue_pos == pos )
               grFillHLine( st->disp_bitmap, 0, blue_pos,
                            st->disp_width, st->blue_color );
@@ -534,8 +570,8 @@
 
 
   static void
-  bitmap_scale( grBitmap*   bit,
-                FT_F26Dot6  scale )
+  bitmap_scale( grBitmap*  bit,
+                int        scale )
   {
     unsigned char*  s = bit->buffer;
     unsigned char*  t;
@@ -640,22 +676,18 @@
                             FTDemo_Handle*   handle,
                             FTDemo_Display*  display )
   {
+    FT_Error      err;
     FT_Size       size;
     FT_GlyphSlot  slot;
-    FT_F26Dot6    scale = st->scale;
+    FT_UInt       glyph_idx;
+    int           scale = st->scale;
     int           ox    = st->x_origin;
     int           oy    = st->y_origin;
 
 
-    if ( st->stroker == NULL )
-    {
-      FT_Stroker_New( handle->library, &st->stroker );
-
-      FT_Stroker_Set( st->stroker, 32, FT_STROKER_LINECAP_BUTT,
-                      FT_STROKER_LINEJOIN_BEVEL, 0x20000 );
-    }
-
-    FTDemo_Get_Size( handle, &size );
+    err = FTDemo_Get_Size( handle, &size );
+    if ( err )
+      return;
 
 #ifdef FT_DEBUG_AUTOFIT
     /* Draw segment before drawing glyph. */
@@ -670,7 +702,7 @@
                            FT_LOAD_NO_BITMAP      |
                            FT_LOAD_FORCE_AUTOHINT |
                            FT_LOAD_TARGET_NORMAL ) )
-        grid_hint_draw_segment( &status, _af_debug_hints );
+        grid_hint_draw_segment( &status, size, _af_debug_hints );
     }
 
     _af_debug_disable_horz_hints = !st->do_horz_hints;
@@ -678,21 +710,41 @@
     _af_debug_disable_blue_hints = !st->do_blue_hints;
 #endif
 
-    if ( FT_Load_Glyph( size->face, (FT_UInt)st->Num,
+    if ( handle->encoding == FT_ENCODING_ORDER )
+      glyph_idx = (FT_UInt)st->Num;
+    else
+      glyph_idx = FTDemo_Get_Index( handle, (FT_UInt32)st->Num );
+
+
+    if ( FT_Load_Glyph( size->face, glyph_idx,
                         handle->load_flags | FT_LOAD_NO_BITMAP ) )
       return;
 
-    /* show advance width */
-    grFillVLine( st->disp_bitmap,
-                 st->x_origin +
-                   ( ( size->face->glyph->metrics.horiAdvance +
-                       size->face->glyph->lsb_delta           -
-                       size->face->glyph->rsb_delta           ) *
-                     st->scale >> 6 ),
-                 0,
-                 st->disp_height,
-                 st->axis_color );
+    if ( st->do_grid )
+    {
+      /* show advance width */
+      grFillVLine( st->disp_bitmap,
+                   st->x_origin +
+                     ( ( size->face->glyph->metrics.horiAdvance +
+                         size->face->glyph->lsb_delta           -
+                         size->face->glyph->rsb_delta           ) *
+                       scale >> 6 ),
+                   0,
+                   st->disp_height,
+                   st->axis_color );
 
+      /* show ascender and descender */
+      grFillHLine( st->disp_bitmap,
+                   0,
+                   st->y_origin - ( size->metrics.ascender  * scale >> 6 ),
+                   st->disp_width,
+                   st->axis_color );
+      grFillHLine( st->disp_bitmap,
+                   0,
+                   st->y_origin - ( size->metrics.descender * scale >> 6 ),
+                   st->disp_width,
+                   st->axis_color );
+    }
 
     slot = size->face->glyph;
     if ( slot->format == FT_GLYPH_FORMAT_OUTLINE )
@@ -981,21 +1033,21 @@
     grWriteln( "Up, Down    adjust size by 0.5pt        if not auto-hinting:                " );
     grWriteln( "                                          H         cycle through hinting   " );
     grWriteln( "Left, Right adjust index by 1                        engines (if available) " );
-    grWriteln( "F7, F8      adjust index by 10          if light auto-hinting:              " );
-    grWriteln( "F9, F10     adjust index by 100           w         toggle warping          " );
-    grWriteln( "F11, F12    adjust index by 1000                      (if available)        " );
+    grWriteln( "F7, F8      adjust index by 16          if normal auto-hinting:             " );
+    grWriteln( "F9, F10     adjust index by 256           w         toggle warping          " );
+    grWriteln( "F11, F12    adjust index by 4096                      (if available)        " );
     grWriteln( "                                                                            " );
     grWriteln( "h           toggle hinting              b           toggle bitmap           " );
     grWriteln( "f           toggle forced auto-         d           toggle dot display      " );
     grWriteln( "             hinting (if hinting)       o           toggle outline display  " );
-    grWriteln( "                                        D           toggle dotnumber display" );
-    grWriteln( "                                                                            " );
-    grWriteln( "F5, F6      cycle through               if Multiple Master or GX font:      " );
-    grWriteln( "             anti-aliasing modes          F2        cycle through axes      " );
-    grWriteln( "L           cycle through LCD             F3, F4    adjust current axis by  " );
-    grWriteln( "             filters                                 1/50th of its range    " );
-    grWriteln( "g, v        adjust gamma value                                              " );
-    grWriteln( "                                        q, ESC      quit ftgrid             " );
+    grWriteln( "G           toggle grid display         D           toggle dotnumber display" );
+    grWriteln( "C           change color palette                                            " );
+    grWriteln( "                                        if Multiple Master or GX font:      " );
+    grWriteln( "F5, F6      cycle through                 F2        cycle through axes      " );
+    grWriteln( "             anti-aliasing modes          F3, F4    adjust current axis by  " );
+    grWriteln( "L           cycle through LCD                        1/50th of its range    " );
+    grWriteln( "             filters                                                        " );
+    grWriteln( "g, v        adjust gamma value          q, ESC      quit ftgrid             " );
     /*          |----------------------------------|    |----------------------------------| */
     grLn();
     grLn();
@@ -1007,36 +1059,154 @@
 
 
   static void
-  event_cff_hinting_engine_change( int  delta )
+  event_font_change( int  delta )
   {
-    int  new_cff_hinting_engine = 0;
+    FT_Error         err;
+    FT_Size          size;
+    FT_UInt          n, num_names;
+    FT_Int           instance_index;
+    FT_Multi_Master  dummy;
+    int              num_indices, is_GX;
 
 
-    if ( delta )
-      new_cff_hinting_engine =
-        ( (int)status.cff_hinting_engine +
-          delta                          +
-          N_CFF_HINTING_ENGINES          ) % N_CFF_HINTING_ENGINES;
+    if ( status.font_index + delta >= handle->num_fonts ||
+         status.font_index + delta < 0                  )
+      return;
 
-    error = FT_Property_Set( handle->library,
-                             "cff",
-                             "hinting-engine",
-                             &new_cff_hinting_engine );
+    status.font_index += delta;
 
-    if ( !error )
+    FTDemo_Set_Current_Font( handle, handle->fonts[status.font_index] );
+    FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
+    FTDemo_Update_Current_Flags( handle );
+
+    num_indices = handle->current_font->num_indices;
+
+    if ( status.Num >= num_indices )
+      status.Num = num_indices - 1;
+
+    err = FTDemo_Get_Size( handle, &size );
+    if ( err )
+      return;
+
+    free( status.mm );
+    status.mm = NULL;
+
+    err = FT_Get_MM_Var( size->face, &status.mm );
+    if ( err )
+      return;
+
+    if ( status.mm->num_axis >= MAX_MM_AXES )
     {
-      /* Resetting the cache is perhaps a bit harsh, but I'm too  */
-      /* lazy to walk over all loaded fonts to check whether they */
-      /* are of type CFF, then unloading them explicitly.         */
-      FTC_Manager_Reset( handle->cache_manager );
-      status.cff_hinting_engine = (FT_UInt)new_cff_hinting_engine;
+      fprintf( stderr, "only handling first %d GX axes (of %d)\n",
+                       MAX_MM_AXES, status.mm->num_axis );
+      status.used_num_axis = MAX_MM_AXES;
+    }
+    else
+      status.used_num_axis = status.mm->num_axis;
+
+    err   = FT_Get_Multi_Master( size->face, &dummy );
+    is_GX = err ? 1 : 0;
+
+    num_names = FT_Get_Sfnt_Name_Count( size->face );
+
+    /* in `face_index', the instance index starts with value 1 */
+    instance_index = ( size->face->face_index >> 16 ) - 1;
+
+    for ( n = 0; n < MAX_MM_AXES; n++ )
+    {
+      free( status.axis_name[n] );
+      status.axis_name[n] = NULL;
     }
 
-    sprintf( status.header_buffer, "CFF engine changed to %s",
-             status.cff_hinting_engine == FT_CFF_HINTING_FREETYPE
-               ? "FreeType" : "Adobe" );
+    for ( n = 0; n < status.used_num_axis; n++ )
+    {
+      if ( status.requested_cnt )
+      {
+        status.design_pos[n] = n < status.requested_cnt
+                                 ? status.requested_pos[n]
+                                 : status.mm->axis[n].def;
+        if ( status.design_pos[n] < status.mm->axis[n].minimum )
+          status.design_pos[n] = status.mm->axis[n].minimum;
+        else if ( status.design_pos[n] > status.mm->axis[n].maximum )
+          status.design_pos[n] = status.mm->axis[n].maximum;
+      }
+      else if ( FT_IS_NAMED_INSTANCE( size->face ) )
+        status.design_pos[n] = status.mm->namedstyle[instance_index].
+                                          coords[n];
+      else
+        status.design_pos[n] = status.mm->axis[n].def;
 
-    status.header = (const char *)status.header_buffer;
+      if ( is_GX )
+      {
+        FT_SfntName  name;
+        FT_UInt      strid, j;
+
+
+        name.string = NULL;
+        strid       = status.mm->axis[n].strid;
+
+        /* iterate over all name entries        */
+        /* to find an English entry for `strid' */
+
+        for ( j = 0; j < num_names; j++ )
+        {
+          error = FT_Get_Sfnt_Name( size->face, j, &name );
+          if ( error )
+            continue;
+
+          if ( name.name_id == strid )
+          {
+            /* XXX we don't have support for Apple's new `ltag' table yet, */
+            /* thus we ignore TT_PLATFORM_APPLE_UNICODE                    */
+            if ( ( name.platform_id == TT_PLATFORM_MACINTOSH &&
+                   name.language_id == TT_MAC_LANGID_ENGLISH )        ||
+                 ( name.platform_id == TT_PLATFORM_MICROSOFT        &&
+                   ( name.language_id & 0xFF )
+                                    == TT_MS_LANGID_ENGLISH_GENERAL ) )
+              break;
+          }
+        }
+
+        if ( name.string )
+        {
+          FT_UInt  len;
+          char*    s;
+
+
+          if ( name.platform_id == TT_PLATFORM_MACINTOSH )
+          {
+            len = put_ascii_string_size( name.string, name.string_len, 0 );
+            s   = (char*)malloc( len );
+            if ( s )
+            {
+              put_ascii_string( s, name.string, name.string_len, 0 );
+              status.axis_name[n] = s;
+            }
+          }
+          else
+          {
+            len = put_unicode_be16_string_size( name.string,
+                                                name.string_len,
+                                                0,
+                                                0 );
+            s   = (char*)malloc( len );
+            if ( s )
+            {
+              put_unicode_be16_string( s,
+                                       name.string,
+                                       name.string_len,
+                                       0,
+                                       0 );
+              status.axis_name[n] = s;
+            }
+          }
+        }
+      }
+    }
+
+    (void)FT_Set_Var_Design_Coordinates( size->face,
+                                         status.used_num_axis,
+                                         status.design_pos );
   }
 
 
@@ -1058,6 +1228,7 @@
       /* lazy to walk over all loaded fonts to check whether they */
       /* are of type TTF, then unloading them explicitly.         */
       FTC_Manager_Reset( handle->cache_manager );
+      event_font_change( 0 );
     }
 
     sprintf( status.header_buffer,
@@ -1072,7 +1243,7 @@
   static void
   event_warping_change( void )
   {
-    if ( handle->lcd_mode == LCD_MODE_LIGHT )
+    if ( handle->lcd_mode == LCD_MODE_AA && handle->autohint )
     {
       FT_Bool  new_warping_state = !status.warping;
 
@@ -1089,34 +1260,28 @@
         /* are auto-hinted, then unloading them explicitly.         */
         FTC_Manager_Reset( handle->cache_manager );
         status.warping = new_warping_state;
+        event_font_change( 0 );
       }
 
       status.header = status.warping ? "warping enabled"
                                      : "warping disabled";
     }
     else
-      status.header = "need light anti-aliasing mode to toggle warping";
+      status.header = "need normal anti-aliasing mode to toggle warping";
   }
 
 
   static void
   event_gamma_change( double  delta )
   {
-    status.gamma += delta;
+    display->gamma += delta;
 
-    if ( status.gamma > 3.0 )
-      status.gamma = 3.0;
-    else if ( status.gamma < 0.0 )
-      status.gamma = 0.0;
+    if ( display->gamma > 3.0 )
+      display->gamma = 3.0;
+    else if ( display->gamma < 0.0 )
+      display->gamma = 0.0;
 
-    grSetGlyphGamma( status.gamma );
-
-    if ( status.gamma == 0.0 )
-      sprintf( status.header_buffer, "gamma changed to sRGB mode" );
-    else
-      sprintf( status.header_buffer, "gamma changed to %.1f", status.gamma );
-
-    status.header = (const char *)status.header_buffer;
+    grSetGlyphGamma( display->gamma );
   }
 
 
@@ -1141,7 +1306,7 @@
   static void
   event_grid_zoom( double  zoom )
   {
-    FT_F26Dot6  scale_old = status.scale;
+    int  scale_old = status.scale;
 
 
     status.scale *= zoom;
@@ -1151,8 +1316,7 @@
     if ( status.scale == scale_old && zoom > 1.0 )
       status.scale++;
 
-    sprintf( status.header_buffer, "zoom level %.0f%%",
-             status.scale * 100.0 / status.scale_0 );
+    sprintf( status.header_buffer, "zoom scale %d:1", status.scale );
 
     status.header = (const char *)status.header_buffer;
   }
@@ -1197,7 +1361,10 @@
     }
 
     if ( delta )
+    {
       FTC_Manager_Reset( handle->cache_manager );
+      event_font_change( 0 );
+    }
 
     sprintf( status.header_buffer, "rendering mode changed to %s",
              lcd_mode );
@@ -1330,156 +1497,6 @@
   }
 
 
-  static void
-  event_font_change( int  delta )
-  {
-    FT_Error         err;
-    FT_Size          size;
-    FT_UInt          n, num_names;
-    FT_Int           instance_index;
-    FT_Multi_Master  dummy;
-    int              num_indices, is_GX;
-
-
-    if ( status.font_index + delta >= handle->num_fonts ||
-         status.font_index + delta < 0                  )
-      return;
-
-    status.font_index += delta;
-
-    FTDemo_Set_Current_Font( handle, handle->fonts[status.font_index] );
-    FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
-    FTDemo_Update_Current_Flags( handle );
-
-    num_indices = handle->current_font->num_indices;
-
-    if ( status.Num >= num_indices )
-      status.Num = num_indices - 1;
-
-    err = FTDemo_Get_Size( handle, &size );
-    if ( err )
-      return;
-
-    free( status.mm );
-    err = FT_Get_MM_Var( size->face, &status.mm );
-    if ( err )
-      return;
-
-    if ( status.mm->num_axis >= MAX_MM_AXES )
-    {
-      fprintf( stderr, "only handling first %d GX axes (of %d)\n",
-                       MAX_MM_AXES, status.mm->num_axis );
-      status.used_num_axis = MAX_MM_AXES;
-    }
-    else
-      status.used_num_axis = status.mm->num_axis;
-
-    err   = FT_Get_Multi_Master( size->face, &dummy );
-    is_GX = err ? 1 : 0;
-
-    num_names = FT_Get_Sfnt_Name_Count( size->face );
-
-    /* in `face_index', the instance index starts with value 1 */
-    instance_index = ( size->face->face_index >> 16 ) - 1;
-
-    for ( n = 0; n < MAX_MM_AXES; n++ )
-    {
-      free( status.axis_name[n] );
-      status.axis_name[n] = NULL;
-    }
-
-    for ( n = 0; n < status.used_num_axis; n++ )
-    {
-      if ( status.requested_cnt )
-      {
-        status.design_pos[n] = n < status.requested_cnt
-                                 ? status.requested_pos[n]
-                                 : status.mm->axis[n].def;
-        if ( status.design_pos[n] < status.mm->axis[n].minimum )
-          status.design_pos[n] = status.mm->axis[n].minimum;
-        else if ( status.design_pos[n] > status.mm->axis[n].maximum )
-          status.design_pos[n] = status.mm->axis[n].maximum;
-      }
-      else if ( FT_IS_NAMED_INSTANCE( size->face ) )
-        status.design_pos[n] = status.mm->namedstyle[instance_index].
-                                          coords[n];
-      else
-        status.design_pos[n] = status.mm->axis[n].def;
-
-      if ( is_GX )
-      {
-        FT_SfntName  name;
-        FT_UInt      strid, j;
-
-
-        name.string = NULL;
-        strid       = status.mm->axis[n].strid;
-
-        /* iterate over all name entries        */
-        /* to find an English entry for `strid' */
-
-        for ( j = 0; j < num_names; j++ )
-        {
-          error = FT_Get_Sfnt_Name( size->face, j, &name );
-          if ( error )
-            continue;
-
-          if ( name.name_id == strid )
-          {
-            /* XXX we don't have support for Apple's new `ltag' table yet, */
-            /* thus we ignore TT_PLATFORM_APPLE_UNICODE                    */
-            if ( ( name.platform_id == TT_PLATFORM_MACINTOSH &&
-                   name.language_id == TT_MAC_LANGID_ENGLISH )        ||
-                 ( name.platform_id == TT_PLATFORM_MICROSOFT        &&
-                   ( name.language_id & 0xFF )
-                                    == TT_MS_LANGID_ENGLISH_GENERAL ) )
-              break;
-          }
-        }
-
-        if ( name.string )
-        {
-          FT_UInt  len;
-          char*    s;
-
-
-          if ( name.platform_id == TT_PLATFORM_MACINTOSH )
-          {
-            len = put_ascii_string_size( name.string, name.string_len, 0 );
-            s   = (char*)malloc( len );
-            if ( s )
-            {
-              put_ascii_string( s, name.string, name.string_len, 0 );
-              status.axis_name[n] = s;
-            }
-          }
-          else
-          {
-            len = put_unicode_be16_string_size( name.string,
-                                                name.string_len,
-                                                0,
-                                                0 );
-            s   = (char*)malloc( len );
-            if ( s )
-            {
-              put_unicode_be16_string( s,
-                                       name.string,
-                                       name.string_len,
-                                       0,
-                                       0 );
-              status.axis_name[n] = s;
-            }
-          }
-        }
-      }
-    }
-
-    (void)FT_Set_Var_Design_Coordinates( size->face,
-                                         status.used_num_axis,
-                                         status.design_pos );
-  }
-
-
   static int
   Process_Event( grEvent*  event )
   {
@@ -1543,6 +1560,20 @@
       break;
 #endif /* FT_DEBUG_AUTOFIT */
 
+    case grKEY( 'C' ):
+      status.do_alt_colors = !status.do_alt_colors;
+      if ( status.do_alt_colors )
+      {
+        status.header = "use alternative colors";
+        grid_status_alt_colors( &status, display );
+      }
+      else
+      {
+        status.header = "use default colors";
+        grid_status_colors( &status, display );
+      }
+      break;
+
     case grKEY( 'L' ):
       event_lcd_filter_change();
       break;
@@ -1565,7 +1596,13 @@
                                       : "glyph hinting is now ignored";
 
       FTC_Manager_Reset( handle->cache_manager );
-      FTDemo_Update_Current_Flags( handle );
+      event_font_change( 0 );
+      break;
+
+    case grKEY( 'G' ):
+      status.do_grid = !status.do_grid;
+      status.header = status.do_grid ? "grid drawing enabled"
+                                     : "grid drawing disabled";
       break;
 
     case grKEY( 'd' ):
@@ -1604,7 +1641,65 @@
           module = &face->driver->root;
 
           if ( !strcmp( module->clazz->module_name, "cff" ) )
-            event_cff_hinting_engine_change( 1 );
+          {
+            if ( FTDemo_Event_Cff_Hinting_Engine_Change(
+                   handle->library,
+                   &status.cff_hinting_engine,
+                   1 ) )
+            {
+              /* Resetting the cache is perhaps a bit harsh, but I'm too  */
+              /* lazy to walk over all loaded fonts to check whether they */
+              /* are of type CFF, then unloading them explicitly.         */
+              FTC_Manager_Reset( handle->cache_manager );
+              event_font_change( 0 );
+            }
+
+            sprintf( status.header_buffer, "CFF engine changed to %s",
+                     status.cff_hinting_engine == FT_HINTING_FREETYPE
+                       ? "FreeType" : "Adobe" );
+
+            status.header = (const char *)status.header_buffer;
+          }
+          else if ( !strcmp( module->clazz->module_name, "type1" ) )
+          {
+            if ( FTDemo_Event_Type1_Hinting_Engine_Change(
+                   handle->library,
+                   &status.type1_hinting_engine,
+                   1 ) )
+            {
+              /* Resetting the cache is perhaps a bit harsh, but I'm too  */
+              /* lazy to walk over all loaded fonts to check whether they */
+              /* are of type Type1, then unloading them explicitly.       */
+              FTC_Manager_Reset( handle->cache_manager );
+              event_font_change( 0 );
+            }
+
+            sprintf( status.header_buffer, "Type 1 engine changed to %s",
+                     status.type1_hinting_engine == FT_HINTING_FREETYPE
+                       ? "FreeType" : "Adobe" );
+
+            status.header = (const char *)status.header_buffer;
+          }
+          else if ( !strcmp( module->clazz->module_name, "t1cid" ) )
+          {
+            if ( FTDemo_Event_T1cid_Hinting_Engine_Change(
+                   handle->library,
+                   &status.t1cid_hinting_engine,
+                   1 ) )
+            {
+              /* Resetting the cache is perhaps a bit harsh, but I'm too  */
+              /* lazy to walk over all loaded fonts to check whether they */
+              /* are of type CID, then unloading them explicitly.         */
+              FTC_Manager_Reset( handle->cache_manager );
+              event_font_change( 0 );
+            }
+
+            sprintf( status.header_buffer, "CID engine changed to %s",
+                     status.t1cid_hinting_engine == FT_HINTING_FREETYPE
+                       ? "FreeType" : "Adobe" );
+
+            status.header = (const char *)status.header_buffer;
+          }
           else if ( !strcmp( module->clazz->module_name, "truetype" ) )
             event_tt_interpreter_version_change();
         }
@@ -1657,14 +1752,14 @@
       break;
 #endif /* FT_DEBUG_AUTOFIT */
 
-    case grKeyLeft:     event_index_change(    -1 ); break;
-    case grKeyRight:    event_index_change(     1 ); break;
-    case grKeyF7:       event_index_change(   -10 ); break;
-    case grKeyF8:       event_index_change(    10 ); break;
-    case grKeyF9:       event_index_change(  -100 ); break;
-    case grKeyF10:      event_index_change(   100 ); break;
-    case grKeyF11:      event_index_change( -1000 ); break;
-    case grKeyF12:      event_index_change(  1000 ); break;
+    case grKeyLeft:     event_index_change(      -1 ); break;
+    case grKeyRight:    event_index_change(       1 ); break;
+    case grKeyF7:       event_index_change(   -0x10 ); break;
+    case grKeyF8:       event_index_change(    0x10 ); break;
+    case grKeyF9:       event_index_change(  -0x100 ); break;
+    case grKeyF10:      event_index_change(   0x100 ); break;
+    case grKeyF11:      event_index_change( -0x1000 ); break;
+    case grKeyF12:      event_index_change(  0x1000 ); break;
 
     case grKeyUp:       event_size_change(  32 ); break;
     case grKeyDown:     event_size_change( -32 ); break;
@@ -1709,95 +1804,28 @@
   static void
   write_header( FT_Error  error_code )
   {
-    FT_Face      face;
-    const char*  basename;
-    const char*  format;
+    FTDemo_Draw_Header( handle, display, status.ptsize, status.res,
+                        status.Num, error_code );
 
-
-    error = FTC_Manager_LookupFace( handle->cache_manager,
-                                    handle->scaler.face_id, &face );
-    if ( error )
-      Fatal( "can't access font file" );
-
-    if ( !status.header )
-    {
-      basename = ft_basename( handle->current_font->filepathname );
-
-      switch ( error_code )
-      {
-      case FT_Err_Ok:
-        sprintf( status.header_buffer, "%.50s %.50s (file `%.100s')",
-                 face->family_name, face->style_name, basename );
-        break;
-
-      case FT_Err_Invalid_Pixel_Size:
-        sprintf( status.header_buffer, "Invalid pixel size (file `%.100s')",
-                 basename );
-        break;
-
-      case FT_Err_Invalid_PPem:
-        sprintf( status.header_buffer, "Invalid ppem value (file `%.100s')",
-                 basename );
-        break;
-
-      default:
-        sprintf( status.header_buffer, "File `%.100s': error 0x%04x",
-                 basename, (FT_UShort)error_code );
-        break;
-      }
-
-      status.header = (const char *)status.header_buffer;
-    }
-
-    grWriteCellString( display->bitmap, 0, 0, status.header,
-                       display->fore_color );
+    if ( status.header )
+      grWriteCellString( display->bitmap, 0, 3 * HEADER_HEIGHT,
+                         status.header, display->fore_color );
 
     if ( status.mm )
     {
-      format = " %s axis: %.02f, %gpt, glyph %d";
+      const char*  format = "%s axis: %.02f";
+
+
       snprintf( status.header_buffer, BUFSIZE, format,
                 status.axis_name[status.current_axis]
                   ? status.axis_name[status.current_axis]
                   : status.mm->axis[status.current_axis].name,
-                status.design_pos[status.current_axis] / 65536.0,
-                status.ptsize / 64.0,
-                status.Num );
+                status.design_pos[status.current_axis] / 65536.0 );
+
+      status.header = (const char *)status.header_buffer;
+      grWriteCellString( display->bitmap, 0, 4 * HEADER_HEIGHT,
+                         status.header, display->fore_color );
     }
-    else
-    {
-      format = " %gpt, glyph %d";
-      snprintf( status.header_buffer, BUFSIZE, format,
-                status.ptsize / 64.0,
-                status.Num );
-    }
-
-    if ( FT_HAS_GLYPH_NAMES( face ) )
-    {
-      char*         p;
-      unsigned int  format_len, gindex, size;
-
-
-      size = strlen( status.header_buffer );
-      p    = status.header_buffer + size;
-      size = BUFSIZE - size;
-
-      format     = ": ";
-      format_len = strlen( format );
-
-      if ( size >= format_len + 2 )
-      {
-        gindex = (unsigned int)status.Num;
-
-        strcpy( p, format );
-        if ( FT_Get_Glyph_Name( face, gindex,
-                                p + format_len, size - format_len ) )
-          *p = '\0';
-      }
-    }
-
-    status.header = (const char *)status.header_buffer;
-    grWriteCellString( display->bitmap, 0, HEADER_HEIGHT,
-                       status.header_buffer, display->fore_color );
 
     grRefreshSurface( display->surface );
   }
@@ -1833,9 +1861,13 @@
     fprintf( stderr,
       "  -r R      Use resolution R dpi (default: 72dpi).\n"
       "  -f index  Specify first index to display (default: 0).\n"
+      "  -e enc    Specify encoding tag (default: no encoding).\n"
+      "            Common values: `unic' (Unicode), `symb' (symbol),\n"
+      "            `ADOB' (Adobe standard), `ADBC' (Adobe custom).\n"
       "  -d \"axis1 axis2 ...\"\n"
       "            Specify the design coordinates for each\n"
-      "            Multiple Master axis at start-up.\n"
+      "            Multiple Master axis at start-up.  Implies `-n'.\n"
+      "  -n        Don't display named instances of variation fonts.\n"
       "\n"
       "  -v        Show version."
       "\n" );
@@ -1856,7 +1888,7 @@
 
     while ( 1 )
     {
-      option = getopt( *argc, *argv, "d:f:h:r:vw:" );
+      option = getopt( *argc, *argv, "d:e:f:h:nr:vw:" );
 
       if ( option == -1 )
         break;
@@ -1878,10 +1910,16 @@
               ++s;
           }
 
-          status.requested_cnt = cnt;
+          status.requested_cnt      = cnt;
+          status.no_named_instances = 1;
         }
         break;
 
+      case 'e':
+        handle->encoding = FTDemo_Make_Encoding_Tag( optarg );
+        status.Num       = 0x20;
+        break;
+
       case 'f':
         status.Num = atoi( optarg );
         break;
@@ -1892,6 +1930,10 @@
           usage( execname );
         break;
 
+      case 'n':
+        status.no_named_instances = 1;
+        break;
+
       case 'r':
         status.res = atoi( optarg );
         if ( status.res < 1 )
@@ -1965,6 +2007,13 @@
     FT_Property_Get( handle->library,
                      "cff",
                      "hinting-engine", &status.cff_hinting_engine );
+    FT_Property_Get( handle->library,
+                     "type1",
+                     "hinting-engine", &status.type1_hinting_engine );
+    FT_Property_Get( handle->library,
+                     "t1cid",
+                     "hinting-engine", &status.t1cid_hinting_engine );
+
 
     /* collect all available versions, then set again the default */
     FT_Property_Get( handle->library,
@@ -1991,12 +2040,18 @@
 
     FT_Library_SetLcdFilter( handle->library, status.lcd_filter );
 
+    FT_Stroker_New( handle->library, &status.stroker );
+
+    FT_Stroker_Set( status.stroker, 32, FT_STROKER_LINECAP_BUTT,
+                      FT_STROKER_LINEJOIN_BEVEL, 0x20000 );
+
     display = FTDemo_Display_New( gr_pixel_mode_rgb24,
                                   status.width, status.height );
     if ( !display )
       Fatal( "could not allocate display surface" );
 
     grid_status_display( &status, display );
+    grid_status_colors(  &status, display );
 
     grSetTitle( display->surface,
                 "FreeType Glyph Grid Viewer - press ? for help" );
@@ -2004,7 +2059,7 @@
     for ( ; argc > 0; argc--, argv++ )
     {
       error = FTDemo_Install_Font( handle, argv[0], 1,
-                                   status.requested_cnt ? 1 : 0 );
+                                   status.no_named_instances ? 1 : 0 );
       if ( error == FT_Err_Invalid_Argument )
         fprintf( stderr, "skipping font `%s' without outlines\n",
                          argv[0] );
@@ -2013,10 +2068,6 @@
     if ( handle->num_fonts == 0 )
       Fatal( "could not find/open any font file" );
 
-    printf( "ptsize =%g\n", status.ptsize / 64.0 );
-    FTDemo_Set_Current_Charsize( handle, status.ptsize, status.res );
-    FTDemo_Update_Current_Flags( handle );
-
     event_font_change( 0 );
 
     grid_status_rescale_initial( &status, handle );
@@ -2025,7 +2076,8 @@
     {
       FTDemo_Display_Clear( display );
 
-      grid_status_draw_grid( &status );
+      if ( status.do_grid )
+        grid_status_draw_grid( &status );
 
       if ( status.work )
         grid_status_draw_outline( &status, handle, display );
@@ -2041,7 +2093,7 @@
 
     for ( n = 0; n < MAX_MM_AXES; n++ )
       free( status.axis_name[n] );
-    free( status.mm );
+    FT_Done_MM_Var( handle->library, status.mm );
 
     FT_Stroker_Done( status.stroker );
     FTDemo_Display_Done( display );
